// g_func.c

#include "g_local.h"


void SUB_CalcMove (Entity *self, Vec3 &tdest, float tspeed, void (*func)(Entity *ent));


#define STATE_TOP		0
#define STATE_BOTTOM	1
#define STATE_UP		2
#define STATE_DOWN		3

#define DOOR_START_OPEN		1
#define DOOR_DONT_LINK		4
#define DOOR_GOLD_KEY		8
#define DOOR_SILVER_KEY		16
#define DOOR_TOGGLE			32


/*

Doors are similar to buttons, but can spawn a fat trigger field around them
to open without a touch, and they link together to form simultanious
double/quad doors.
 
Door.owner is the master door.  If there is only one door, it points to itself.
If multiple doors, all will point to a single one.

Door.enemy chains from the master door through all doors linked in the chain.

*/

/*
=============================================================================

THINK FUNCTIONS

=============================================================================
*/

void door_go_down (Entity *self);
void door_go_up (Entity *self, Entity *activator);

void door_blocked (Entity *self, Entity *other)
{
	G_Damage (other, self, self->goalentity, self->dmg, DT_SQUISH);

// if a door has a negative wait, it would never come back if blocked,
// so let it just squash the object to death real fast
	if (self->wait >= 0)
	{
		if (self->state == STATE_DOWN)
			door_go_up (self, other);
		else
			door_go_down (self);
	}
}


void door_hit_top (Entity *self)
{
	G_Sound (self, CHAN_NO_PHS_ADD+CHAN_VOICE, self->noise1);
	self->state = STATE_TOP;
	if (self->spawnflags & DOOR_TOGGLE)
		return;         // don't come down automatically
	self->think = door_go_down;
	self->nextthink = self->ltime + self->wait;
}

void door_hit_bottom (Entity *self)
{
	G_Sound (self, CHAN_NO_PHS_ADD+CHAN_VOICE, self->noise1);
	self->state = STATE_BOTTOM;
}

void door_go_down (Entity *self)
{
	G_Sound (self, CHAN_VOICE, self->noise2);
	if (self->max_health)
	{
		self->takedamage = DAMAGE_YES;
		self->health = self->max_health;
	}
	
	self->state = STATE_DOWN;
	SUB_CalcMove (self, self->pos1, self->speed, door_hit_bottom);
}

void door_go_up (Entity *self, Entity *activator)
{
	if (self->state == STATE_UP)
		return;         // already going up

	if (self->state == STATE_TOP)
	{       // reset top wait time
		self->nextthink = self->ltime + self->wait;
		return;
	}
	
	G_Sound (self, CHAN_VOICE, self->noise2);
	self->state = STATE_UP;
	SUB_CalcMove (self, self->pos2, self->speed, door_hit_top);

	G_UseTargets (self, activator);
}


/*
=============================================================================

ACTIVATION FUNCTIONS

=============================================================================
*/

void door_fire (Entity *self, Entity *activator)
{
	Entity	*ent;

// FIXME: happens on e1m7
//	if (self->owner != self)
//		objerror ("door_fire: self->owner != self");

// play use key sound

	if (self->items)
		G_Sound (self, CHAN_VOICE, self->noise4);

	self->message = "";             // no more message

	if (self->spawnflags & DOOR_TOGGLE)
	{
		if (self->state == STATE_UP || self->state == STATE_TOP)
		{
			ent = self;
			do
			{
				door_go_down (ent);
				ent = ent->enemy;
			} while (ent != self && ent && ent != world);
			return;
		}
	}
	
// trigger all paired doors
	ent = self;
	do
	{
		ent->goalentity = activator;		// Who fired us
		door_go_up (ent, activator);
		ent = ent->enemy;
	} while (ent != self && ent && ent != world);
}


void door_use (Entity *self, Entity *other, Entity *activator)
{
	self->message = "";			// door messages are for touch only
	if (self->owner)
		self->owner->message = "";
	if (self->enemy)
		self->enemy->message = "";

	door_fire (self, activator);
}


void door_trigger_touch (Entity *self, Entity *other)
{
	if (other->health <= 0)
		return;

	if (g_time < self->attack_finished)
		return;
	self->attack_finished = g_time + 1;

	door_use (self->owner, other, other);
}


void door_killed (Entity *self, Entity *inflictor, Entity *attacker)
{
	Entity *ent;
	
	ent = self->owner;
	ent->health = ent->max_health;
	ent->takedamage = DAMAGE_NO;    // will be reset upon return
	door_use (ent, attacker, attacker);
}


/*
================
door_touch

Prints messages and opens key doors
================
*/
void door_touch (Entity *self, Entity *other)
{
	if (!other->player)
		return;

	if (self->owner->attack_finished > g_time)
		return;

	self->owner->attack_finished = g_time + 2;

	if (self->owner->message && self->owner->message[0])
	{
		G_Centerprint (other, self->owner->message);
		G_Sound (other, CHAN_VOICE, "misc/talk.wav");
	}
	
// key door stuff
	if (!self->items)
		return;

// FIXME: blink key on player's status bar
	if ( (self->items & other->items) != self->items )
	{
		if (self->owner->items == IT_KEY1)
		{
			if (world->worldtype == 2)
				G_Centerprint (other, "You need the silver keycard");
			else if (world->worldtype == 1)
				G_Centerprint (other, "You need the silver runekey");
			else
				G_Centerprint (other, "You need the silver key");
			G_Sound (self, CHAN_VOICE, self->noise3);
		}
		else
		{
			if (world->worldtype == 2)
				G_Centerprint (other, "You need the gold keycard");
			else if (world->worldtype == 1)
				G_Centerprint (other, "You need the gold runekey");
			else
				G_Centerprint (other, "You need the gold key");
			G_Sound (self, CHAN_VOICE, self->noise3);
		}
		return;
	}

	other->items = other->items - self->items;
	self->touch = NULL;
	if (self->enemy)
		self->enemy->touch = NULL;    // get paired door
	door_use (self, other, other);	// FIXME: self, other, other - right?
}

/*
=============================================================================

SPAWNING FUNCTIONS

=============================================================================
*/


Entity *spawn_field (Entity *self, Vec3& fmins, Vec3& fmaxs)
{
	Entity *trigger = G_Spawn();
	trigger->movetype = MOVETYPE_NONE;
	trigger->solid = SOLID_TRIGGER;
	trigger->owner = self;
	trigger->touch = door_trigger_touch;

	setsize (trigger, fmins[0] - 60, fmins[1] - 60, fmins[2] - 8,
					fmaxs[0] + 60, fmaxs[1] + 60, fmaxs[2] + 8);

	G_LinkEntity (trigger);
	return trigger;
}


qboolean EntitiesTouching (Entity *e1, Entity *e2)
{
	if (e1->mins[0] > e2->maxs[0])
		return false;
	if (e1->mins[1] > e2->maxs[1])
		return false;
	if (e1->mins[2] > e2->maxs[2])
		return false;
	if (e1->maxs[0] < e2->mins[0])
		return false;
	if (e1->maxs[1] < e2->mins[1])
		return false;
	if (e1->maxs[2] < e2->mins[2])
		return false;
	return true;
}


/*
=============
LinkDoors


=============
*/
void LinkDoors (Entity *self)
{
	Entity		*t, *starte;
	Vec3		cmins, cmaxs;

	if (self->enemy)
		return;         // already linked by another door
	if (self->spawnflags & 4)
	{
		self->owner = self->enemy = self;
		return;         // don't want to link this door
	}

	cmins = self->mins;
	cmaxs = self->maxs;
	
	starte = self;
	t = self;
	
	while (1)
	{
		self->owner = starte;					// master door

		if (self->health)
			starte->health = self->health;
		if (self->targetname)
			starte->targetname = self->targetname;
		if (self->message && self->message[0])
			starte->message = self->message;

		t = G_Find (t, FOFS(classname), self->classname);
		if (!t)
		{
			self->enemy = starte;				// make the chain a loop

		// shootable, fired, or key doors just needed the owner/enemy links,
		// they don't spawn a field
	
			self = self->owner;

			if (self->health)
				return;
			if (self->targetname)
				return;
			if (self->items)
				return;

			self->owner->trigger_field = spawn_field(self, cmins, cmaxs);

			return;
		}

		if (EntitiesTouching(self,t))
		{
			if (t->enemy)
				objerror ("cross connected doors");
			
			self->enemy = t;
			self = t;

			if (t->mins[0] < cmins[0])
				cmins[0] = t->mins[0];
			if (t->mins[1] < cmins[1])
				cmins[1] = t->mins[1];
			if (t->mins[2] < cmins[2])
				cmins[2] = t->mins[2];
			if (t->maxs[0] > cmaxs[0])
				cmaxs[0] = t->maxs[0];
			if (t->maxs[1] > cmaxs[1])
				cmaxs[1] = t->maxs[1];
			if (t->maxs[2] > cmaxs[2])
				cmaxs[2] = t->maxs[2];
		}
	}
}


/*QUAKED func_door (0 .5 .8) ? START_OPEN x DOOR_DONT_LINK GOLD_KEY SILVER_KEY TOGGLE
if two doors touch, they are assumed to be connected and operate as a unit.

TOGGLE causes the door to wait in both the start and end states for a trigger event.

START_OPEN causes the door to move to its destination when spawned, and operate in reverse.  It is used to temporarily or permanently close off an area when triggered (not usefull for touch or takedamage doors).

Key doors are always wait -1.

"message"       is printed when the door is touched if it is a trigger door and it hasn't been fired yet
"angle"         determines the opening direction
"targetname" if set, no touch field will be spawned and a remote button or trigger field activates the door.
"health"        if set, door must be shot open
"speed"         movement speed (100 default)
"wait"          wait before returning (3 default, -1 = never return)
"lip"           lip remaining at end of move (8 default)
"dmg"           damage to inflict when blocked (2 default)
"sounds"
0)      no sound
1)      stone
2)      base
3)      stone chain
4)      screechy metal
*/

void SP_func_door (Entity *ent)
{
	if (world->worldtype == 0) {
		G_PrecacheSound ("doors/medtry.wav");
		G_PrecacheSound ("doors/meduse.wav");
		ent->noise3 = "doors/medtry.wav";
		ent->noise4 = "doors/meduse.wav";
	}
	else if (world->worldtype == 1) {
		G_PrecacheSound ("doors/runetry.wav");
		G_PrecacheSound ("doors/runeuse.wav");
		ent->noise3 = "doors/runetry.wav";
		ent->noise4 = "doors/runeuse.wav";
	}
	else if (world->worldtype == 2) {
		G_PrecacheSound ("doors/basetry.wav");
		G_PrecacheSound ("doors/baseuse.wav");
		ent->noise3 = "doors/basetry.wav";
		ent->noise4 = "doors/baseuse.wav";
	}
	else {
		gi.dprintf ("no worldtype set!\n");
	}

	if (ent->sounds == 1) {
		G_PrecacheSound ("doors/drclos4.wav");
		G_PrecacheSound ("doors/doormv1.wav");
		ent->noise1 = "doors/drclos4.wav";
		ent->noise2 = "doors/doormv1.wav";
	}
	else if (ent->sounds == 2) {
		G_PrecacheSound ("doors/hydro1.wav");
		G_PrecacheSound ("doors/hydro2.wav");
		ent->noise2 = "doors/hydro1.wav";
		ent->noise1 = "doors/hydro2.wav";
	}
	else if (ent->sounds == 3) {
		G_PrecacheSound ("doors/stndr1.wav");
		G_PrecacheSound ("doors/stndr2.wav");
		ent->noise2 = "doors/stndr1.wav";
		ent->noise1 = "doors/stndr2.wav";
	}
	else if (ent->sounds == 4) {
		G_PrecacheSound ("doors/ddoor1.wav");
		G_PrecacheSound ("doors/ddoor2.wav");
		ent->noise1 = "doors/ddoor2.wav";
		ent->noise2 = "doors/ddoor1.wav";
	}
	else {
		// no sounds
	}

	G_SetMovedir (ent->angles, ent->movedir);

	ent->max_health = ent->health;
	ent->solid = SOLID_BSP;
	ent->movetype = MOVETYPE_PUSH;
	ent->classname = "door";
	gi.setmodel (ent, ent->model);
//	G_LinkEntity (ent);

	ent->blocked = door_blocked;
	ent->use = door_use;
	
	if (ent->spawnflags & DOOR_SILVER_KEY)
		ent->items = IT_KEY1;
	if (ent->spawnflags & DOOR_GOLD_KEY)
		ent->items = IT_KEY2;
	
	if (!ent->speed)
		ent->speed = 100;
	if (!ent->wait)
		ent->wait = 3;
	if (!ent->lip)
		ent->lip = 8;
	if (!ent->dmg)
		ent->dmg = 2;

	Vec3 size = ent->maxs - ent->mins;

	ent->pos1 = ent->origin;
	ent->pos2 = ent->pos1 +
		ent->movedir * ( fabs(ent->movedir * size) - ent->lip );

// DOOR_START_OPEN is to allow an entity to be lighted in the closed position
// but spawn in the open position
	if (ent->spawnflags & DOOR_START_OPEN)
	{
		ent->origin = ent->pos2;
		G_LinkEntity (ent);
		ent->pos2 = ent->pos1;
		ent->pos1 = ent->origin;
	}

	ent->state = STATE_BOTTOM;

	if (ent->health)
	{
		ent->takedamage = DAMAGE_YES;
		ent->die = door_killed;
	}
	
	if (ent->items)
		ent->wait = -1;
		
	ent->touch = door_touch;

// LinkDoors can't be done until all of the doors have been spawned, so
// the sizes can be detected properly.
	ent->think = LinkDoors;
	ent->nextthink = ent->ltime + 0.1;
}


/*
=============================================================================

SECRET DOORS

=============================================================================
*/

#define SECRET_OPEN_ONCE	1		// stays open
#define SECRET_1ST_LEFT 	2		// 1st move is left of arrow
#define SECRET_1ST_DOWN 	4		// 1st move is down from arrow
#define SECRET_NO_SHOOT 	8		// only opened by trigger
#define SECRET_YES_SHOOT	16		// shootable even if targeted

void door_secret_move1 (Entity *self);
void door_secret_move2 (Entity *self);
void door_secret_move3 (Entity *self);
void door_secret_move4 (Entity *self);
void door_secret_move5 (Entity *self);
void door_secret_move6 (Entity *self);
void door_secret_done (Entity *self);
void door_secret_use (Entity *self);
void door_secret_pain (Entity *self, Entity *other);

// Wait after first movement...
void door_secret_move1 (Entity *self)
{
	self->nextthink = self->ltime + 1.0;
	self->think = door_secret_move2;
	G_Sound (self, CHAN_VOICE, self->noise3);
}

// Start moving sideways w/sound...
void door_secret_move2 (Entity *self)
{
	G_Sound (self, CHAN_VOICE, self->noise2);
	SUB_CalcMove (self, self->dest2, self->speed, door_secret_move3);
}

// Wait here until time to go back...
void door_secret_move3 (Entity *self)
{
	G_Sound (self, CHAN_VOICE, self->noise3);
	if (!(self->spawnflags & SECRET_OPEN_ONCE))
	{
		self->nextthink = self->ltime + self->wait;
		self->think = door_secret_move4;
	}
}

// Move backward...
void door_secret_move4 (Entity *self)
{
	G_Sound (self, CHAN_VOICE, self->noise2);
	SUB_CalcMove (self, self->dest1, self->speed, door_secret_move5);          
}

// Wait 1 second...
void door_secret_move5 (Entity *self)
{
	self->nextthink = self->ltime + 1.0;
	self->think = door_secret_move6;
	G_Sound (self, CHAN_VOICE, self->noise3);
}

void door_secret_move6 (Entity *self)
{
	G_Sound(self, CHAN_VOICE, self->noise2);
	SUB_CalcMove (self, self->oldorigin, self->speed, door_secret_done);
}

void door_secret_done (Entity *self)
{
	if (!self->targetname || self->spawnflags&SECRET_YES_SHOOT)
	{
		self->health = 10000;
		self->takedamage = DAMAGE_YES;
		self->pain = door_secret_pain;
//@@@		self->die = door_secret_use;
	}
	G_Sound (self, CHAN_NO_PHS_ADD+CHAN_VOICE, self->noise3);
}


void door_secret_use (Entity *self, Entity *other, Entity *activator)
{
	self->health = 10000;

	// exit if still moving around...
	if (self->origin != self->oldorigin)
		return;
	
	self->message = "";             // no more message

	G_UseTargets (self, activator);
	
	if (!(self->spawnflags & SECRET_NO_SHOOT))
	{
		self->pain = NULL;
		self->takedamage = DAMAGE_NO;
	}
	self->velocity.Clear();

	// Make a sound, wait a little...
	
	G_Sound (self, CHAN_VOICE, self->noise1);
	self->nextthink = self->ltime + 0.1;

	Vec3 forward, right, up;
	AngleVectors (self->mangle, forward, right, up);
	
	Vec3 size = self->maxs - self->mins;

	if (!self->t_width)
	{
		if (self->spawnflags & SECRET_1ST_DOWN)
			self->t_width = fabs(up * size);
		else
			self->t_width = fabs(right * size);
	}
		
	if (!self->t_length)
		self->t_length = fabs(forward * size);

	if (self->spawnflags & SECRET_1ST_DOWN)
		self->dest1 = self->origin - up * self->t_width;
	else {
		int temp = (self->spawnflags & SECRET_1ST_LEFT) ? -1 : 1;
		self->dest1 = self->origin + right * (self->t_width * temp);
	}
		
	self->dest2 = self->dest1 + forward * self->t_length;
	SUB_CalcMove (self, self->dest1, self->speed, door_secret_move1);
	G_Sound (self, CHAN_VOICE, self->noise2);
}

void door_secret_pain (Entity *self, Entity *other)
{
	door_secret_use (self, other, other);
}

void door_secret_touch (Entity *self, Entity *other)
{
	if (!other->player)
		return;
	if (self->attack_finished > g_time)
		return;

	self->attack_finished = g_time + 2;
	
	if (self->message && self->message[0]) {
		gi.centerprint (other, self->message);
		G_Sound (other, CHAN_BODY, "misc/talk.wav");
	}
}

void door_secret_blocked (Entity *self, Entity *other)
{
	if (g_time < self->attack_finished)
		return;
	self->attack_finished = g_time + 0.5;
	G_Damage (other, self, self, self->dmg, DT_SQUISH);
}


/*QUAKED func_door_secret (0 .5 .8) ? open_once 1st_left 1st_down no_shoot always_shoot
Basic secret door. Slides back, then to the side. Angle determines direction.
wait  = # of seconds before coming back
1st_left = 1st move is left of arrow
1st_down = 1st move is down from arrow
always_shoot = even if targeted, keep shootable
t_width = override WIDTH to move back (or height if going down)
t_length = override LENGTH to move sideways
"dmg"           damage to inflict when blocked (2 default)

If a secret door has a targetname, it will only be opened by its button or trigger, not by damage.
"sounds"
1) medieval
2) metal
3) base
*/
void SP_func_door_secret (Entity *ent)
{
	if (!ent->sounds)
		ent->sounds = 3;

	if (ent->sounds == 1) {
		G_PrecacheSound ("doors/latch2.wav");
		G_PrecacheSound ("doors/winch2.wav");
		G_PrecacheSound ("doors/drclos4.wav");
		ent->noise1 = "doors/latch2.wav";
		ent->noise2 = "doors/winch2.wav";
		ent->noise3 = "doors/drclos4.wav";
	}
	else if (ent->sounds == 2) {
		G_PrecacheSound ("doors/airdoor1.wav");
		G_PrecacheSound ("doors/airdoor2.wav");
		ent->noise2 = "doors/airdoor1.wav";
		ent->noise1 = "doors/airdoor2.wav";
		ent->noise3 = "doors/airdoor2.wav";
	}
	else if (ent->sounds == 3) {
		G_PrecacheSound ("doors/basesec1.wav");
		G_PrecacheSound ("doors/basesec2.wav");
		ent->noise2 = "doors/basesec1.wav";
		ent->noise1 = "doors/basesec2.wav";
		ent->noise3 = "doors/basesec2.wav";
	}

	if (!ent->dmg)
		ent->dmg = 2;
		
	// Magic formula...
	ent->mangle = ent->angles;
	ent->angles.Clear();
	ent->solid = SOLID_BSP;
	ent->movetype = MOVETYPE_PUSH;
	ent->classname = "door";
	gi.setmodel (ent, ent->model);
	G_LinkEntity (ent);
	
	ent->touch = door_secret_touch;
	ent->blocked = door_secret_blocked;
	ent->use = door_secret_use;
	if (!ent->targetname || ent->spawnflags & SECRET_YES_SHOOT)
	{
		ent->health = 10000;
		ent->takedamage = DAMAGE_YES;
		ent->pain = door_secret_pain;
	}
	ent->oldorigin = ent->origin;		// why?

	if (!ent->speed)
		ent->speed = 50;
	if (!ent->wait)
		ent->wait = 5;          // 5 seconds before closing
}


/*
=============
SUB_CalcMove

calculate self->velocity and self->nextthink to reach dest from
self->origin traveling at speed
===============
*/

// After moving, set origin to exact final destination
void SUB_CalcMoveDone (Entity *self)
{
	self->origin = self->finaldest;
	G_LinkEntity (self);
	self->velocity.Clear();
	self->nextthink = -1;
	if (self->think1)
		self->think1 (self);
}

void SUB_CalcMove (Entity *self, Vec3 &tdest, float tspeed, void (*func)(Entity *ent))
{
	Vec3	vdestdelta;
	float	len, traveltime;

	if (!tspeed)
		objerror("No speed is defined!");

	self->think1 = func;
	self->finaldest = tdest;
	self->think = SUB_CalcMoveDone;

	if (tdest == self->origin)
	{
		self->velocity.Clear();
		self->nextthink = self->ltime + 0.1;
		return;
	}
		
// set destdelta to the vector needed to move
	vdestdelta = tdest - self->origin;
	
// calculate length of vector
	len = Length(vdestdelta);
	
// divide by speed to get time to reach dest
	traveltime = len / tspeed;

	if (traveltime < 0.03)
		traveltime = 0.03;
	
// set nextthink to trigger a think when dest is reached
	self->nextthink = self->ltime + traveltime;

// scale the destdelta vector by the time spent traveling to get velocity
	self->velocity = vdestdelta * (1/traveltime);
}


//============================================================================

void plat_center_touch (Entity *self, Entity *other);
void plat_outside_touch (Entity *self, Entity *other);
void plat_trigger_use (Entity *self, Entity *other, Entity *activator);
void plat_go_up (Entity *self);
void plat_go_down (Entity *self);
void plat_blocked (Entity *self, Entity *other);
#define PLAT_LOW_TRIGGER	1

void plat_spawn_inside_trigger (Entity *self)
{
	Entity	*trigger;
	Vec3	tmin, tmax;

//
// middle trigger
//	
	trigger = G_Spawn();
	trigger->touch = plat_center_touch;
	trigger->movetype = MOVETYPE_NONE;
	trigger->solid = SOLID_TRIGGER;
	trigger->enemy = self;
	
	tmin[0] = self->mins[0] + 25;
	tmin[1] = self->mins[1] + 25;
	tmin[2] = self->mins[2];

	tmax[0] = self->maxs[0] - 25;
	tmax[1] = self->maxs[1] - 25;
	tmax[2] = self->maxs[2] + 8;

	tmin[2] = tmax[2] - (self->pos1[2] - self->pos2[2] + 8);

	if (self->spawnflags & PLAT_LOW_TRIGGER)
		tmax[2] = tmin[2] + 8;
	
	if (self->maxs[0] - self->mins[0] <= 50)
	{
		tmin[0] = (self->mins[0] + self->maxs[0]) / 2;
		tmax[0] = tmin[0] + 1;
	}
	if (self->maxs[1] - self->mins[1] <= 50)
	{
		tmin[1] = (self->mins[1] + self->maxs[1]) / 2;
		tmax[1] = tmin[1] + 1;
	}
	
	setsize (trigger, tmin[0], tmin[1], tmin[2], tmax[0], tmax[1], tmax[2]);

	G_LinkEntity (self);
}

void plat_hit_top (Entity *self)
{
	G_Sound (self, CHAN_NO_PHS_ADD+CHAN_VOICE, self->noise1);
	self->state = STATE_TOP;
	self->think = plat_go_down;
	self->nextthink = self->ltime + 3;
}

void plat_hit_bottom (Entity *self)
{
	G_Sound (self, CHAN_NO_PHS_ADD+CHAN_VOICE, self->noise1);
	self->state = STATE_BOTTOM;
}

void plat_go_down (Entity *self)
{
	G_Sound (self, CHAN_VOICE, self->noise);
	self->state = STATE_DOWN;
	SUB_CalcMove (self, self->pos2, self->speed, plat_hit_bottom);
}

void plat_go_up (Entity *self)
{
	G_Sound (self, CHAN_VOICE, self->noise);
	self->state = STATE_UP;
	SUB_CalcMove (self, self->pos1, self->speed, plat_hit_top);
}

void plat_center_touch (Entity *self, Entity *other)
{
	if (!other->player)
		return;

	if (other->health <= 0)
		return;

	self = self->enemy;
	if (self->state == STATE_BOTTOM)
		plat_go_up (self);
	else if (self->state == STATE_TOP)
		self->nextthink = self->ltime + 1;	// delay going down
}

void plat_outside_touch (Entity *self, Entity *other)
{
	if (!other->player)
		return;

	if (other->health <= 0)
		return;
		
	self = self->enemy;
	if (self->state == STATE_TOP)
		plat_go_down (self);
}

void plat_trigger_use (Entity *self, Entity *other, Entity *activator)
{
	if (self->think)
		return;		// already activated
	plat_go_down (self);
}


void plat_blocked (Entity *self, Entity *other)
{
	G_Damage (other, self, self, 1, DT_SQUISH);
	
	if (self->state == STATE_UP)
		plat_go_down (self);
	else if (self->state == STATE_DOWN)
		plat_go_up (self);
	else
		objerror ("plat_blocked: bad self->state\n");
}

void plat_use (Entity *self, Entity *other, Entity *activator)
{
	self->use = NULL;
	if (self->state != STATE_UP)
		objerror ("plat_use: not in up state");
	plat_go_down (self);
}


/*QUAKED func_plat (0 .5 .8) ? PLAT_LOW_TRIGGER
speed	default 150

Plats are always drawn in the extended position, so they will light correctly.

If the plat is the target of another trigger or button, it will start out disabled in the extended position until it is trigger, when it will lower and become a normal plat.

If the "height" key is set, that will determine the amount the plat moves, instead of being implicitly determined by the model's height.
Set "sounds" to one of the following:
1) base fast
2) chain slow
*/
void SP_func_plat (Entity *ent)
{
//	if (!ent->t_length)
//		ent->t_length = 80;
//	if (!ent->t_width)
//		ent->t_width = 10;

	if (ent->sounds == 0)
		ent->sounds = 2;
// FIX THIS TO LOAD A GENERIC PLAT SOUND

	if (ent->sounds == 1)
	{
		G_PrecacheSound ("plats/plat1.wav");
		G_PrecacheSound ("plats/plat2.wav");
		ent->noise = "plats/plat1.wav";
		ent->noise1 = "plats/plat2.wav";
	}

	if (ent->sounds == 2)
	{
		G_PrecacheSound ("plats/medplat1.wav");
		G_PrecacheSound ("plats/medplat2.wav");
		ent->noise = "plats/medplat1.wav";
		ent->noise1 = "plats/medplat2.wav";
	}


	ent->mangle = ent->angles;
	ent->angles.Clear();
	
	ent->classname = "plat";
	ent->solid = SOLID_BSP;
	ent->movetype = MOVETYPE_PUSH;
	gi.setmodel (ent, ent->model);

	ent->blocked = plat_blocked;
	if (!ent->speed)
		ent->speed = 150;

// pos1 is the top position, pos2 is the bottom
	ent->pos1 = ent->origin;
	ent->pos2 = ent->origin;
	if (ent->height)
		ent->pos2[2] -= ent->height;
	else
//		ent->pos2[2] -= (ent->maxs[2] - ent->mins[2]) - ent->lip; // the right way
		ent->pos2[2] -= (ent->maxs[2] - ent->mins[2]) - 8;

	ent->use = plat_trigger_use;

	plat_spawn_inside_trigger (ent);	// the "start moving" trigger	

	if (ent->targetname)
	{
		ent->state = STATE_UP;
		ent->use = plat_use;
	}
	else
	{
		ent->origin = ent->pos2;
		ent->state = STATE_BOTTOM;
	}

	G_LinkEntity (ent);
}

//============================================================================

void train_next (Entity *self);
void func_train_find (Entity *self);

void train_blocked (Entity *self, Entity *other)
{
	if (g_time < self->attack_finished)
		return;
	self->attack_finished = g_time + 0.5;
	G_Damage (other, self, self, self->dmg, DT_SQUISH);
}

void train_use (Entity *self, Entity *other, Entity *activator)
{
	if (self->think != func_train_find)
		return;		// already activated
	train_next (self);
}

void train_wait (Entity *self)
{
	if (self->wait)
	{
		self->nextthink = self->ltime + self->wait;
		G_Sound (self, CHAN_NO_PHS_ADD+CHAN_VOICE, self->noise);
	}
	else
		self->nextthink = self->ltime + 0.1;
	
	self->think = train_next;
}

void train_next (Entity *self)
{
	Entity *targ;
	Vec3	tdest;

	if (!self->target)
		objerror ("train_next: no next target");

	targ = G_Find (NULL, FOFS(targetname), self->target);
	if (!targ)
		objerror ("train_next: bad target %s", self->target);

	self->target = targ->target;

	if (targ->wait)
		self->wait = targ->wait;
	else
		self->wait = 0;

	G_Sound (self, CHAN_VOICE, self->noise1);
	tdest = targ->origin - self->mins;
	SUB_CalcMove (self, tdest, self->speed, train_wait);
}

void func_train_find (Entity *self)
{
	if (!self->target) {
		gi.dprintf ("func_train_find: no target\n");
		return;
	}

	Entity *targ = G_Find (NULL, FOFS(targetname), self->target);
	if (!targ) {
		gi.dprintf ("func_train_find: target %s not found\n", self->target);
		return;
	}

	self->target = targ->target;
	self->origin = targ->origin - self->mins;
	G_LinkEntity (self);

	if (!self->targetname) {
		// not triggered, so start immediately
		self->nextthink = self->ltime + 0.1;
		self->think = train_next;
	}
}


/*QUAKED func_train (0 .5 .8) ?
Trains are moving platforms that players can ride.
The targets origin specifies the min point of the train at each corner.
The train spawns at the first target it is pointing at.
If the train is the target of a button or trigger, it will not begin moving until activated.
speed	default 100
dmg		default	2
sounds
1) ratchet metal
*/
void SP_func_train (Entity *ent)
{	
	if (!ent->speed)
		ent->speed = 100;
	if (!ent->target)
		objerror ("func_train without a target");
	if (!ent->dmg)
		ent->dmg = 2;

	if (ent->sounds == 1)
	{
		G_PrecacheSound ("plats/train2.wav");
		G_PrecacheSound ("plats/train1.wav");
		ent->noise = ("plats/train2.wav");
		ent->noise1 = ("plats/train1.wav");
	}

	ent->solid = SOLID_BSP;
	ent->movetype = MOVETYPE_PUSH;
	ent->blocked = train_blocked;
	ent->use = train_use;
	ent->classname = "train";

	gi.setmodel (ent, ent->model);
	G_LinkEntity (ent);

// start trains on the second frame, to make sure their targets have had
// a chance to spawn
	ent->nextthink = ent->ltime + 0.1;
	ent->think = func_train_find;
}

/*QUAKED misc_teleporttrain (0 .5 .8) (-8 -8 -8) (8 8 8)
This is used for the final bos
*/
void SP_misc_teleporttrain (Entity *ent)
{	
	if (!ent->speed)
		ent->speed = 100;
	if (!ent->target)
		objerror ("func_train without a target");

	ent->solid = SOLID_NOT;
	ent->movetype = MOVETYPE_PUSH;
	ent->blocked = train_blocked;
	ent->use = train_use;
	ent->avelocity = Angles(200, 100, 300);

	G_PrecacheModel ("progs/teleport.mdl");
	gi.setmodel (ent, "progs/teleport.mdl");
	G_LinkEntity (ent);

// start trains on the second frame, to make sure their targets have had
// a chance to spawn
	ent->nextthink = ent->ltime + 0.1;
	ent->think = func_train_find;
}

/*
======================================================================

BUTTONS

======================================================================
*/

void button_wait (Entity *self);
void button_return (Entity *self);
void button_done (Entity *self);

void button_fire (Entity *self)
{
	if (self->state == STATE_UP || self->state == STATE_TOP)
		return;

	G_Sound (self, CHAN_VOICE, self->noise);

	self->state = STATE_UP;
	SUB_CalcMove (self, self->pos2, self->speed, button_wait);
}

void button_wait (Entity *self)
{
	self->state = STATE_TOP;
	self->nextthink = self->ltime + self->wait;
	self->think = button_return;
	G_UseTargets (self, self->enemy);
	self->frame = 1;			// use alternate textures
}

void button_return (Entity *self)
{
	self->state = STATE_DOWN;
	SUB_CalcMove (self, self->pos1, self->speed, button_done);
	self->frame = 0;			// use normal textures
	if (self->health)
		self->takedamage = DAMAGE_YES;	// can be shot again
}

void button_done (Entity *self)
{
	self->state = STATE_BOTTOM;
}


void button_blocked (Entity *self, Entity *other)
{	// do nothing, just don't come all the way back out
}


void button_use (Entity *self, Entity *other, Entity *activator)
{
	self->enemy = activator;
	button_fire (self);
}

void button_touch (Entity *self, Entity *other)
{
	if (!other->player)
		return;
	self->enemy = other;
	button_fire (self);
}

void button_killed (Entity *self, Entity *inflictor, Entity *attacker)
{
	self->enemy = attacker;
	self->health = self->max_health;
	self->takedamage = DAMAGE_NO;	// will be reset upon return
	button_fire (self);
}


/*QUAKED func_button (0 .5 .8) ?
When a button is touched, it moves some distance in the direction of it's angle, triggers all of it's targets, waits some time, then returns to it's original position where it can be triggered again.

"angle"		determines the opening direction
"target"	all entities with a matching targetname will be used
"speed"		override the default 40 speed
"wait"		override the default 1 second wait (-1 = never return)
"lip"		override the default 4 pixel lip remaining at end of move
"health"	if set, the button must be killed instead of touched
"sounds"
0) steam metal
1) wooden clunk
2) metallic click
3) in-out
*/
void SP_func_button (Entity *self)
{
	if (self->sounds == 0)
	{
		G_PrecacheSound ("buttons/airbut1.wav");
		self->noise = "buttons/airbut1.wav";
	}
	if (self->sounds == 1)
	{
		G_PrecacheSound ("buttons/switch21.wav");
		self->noise = "buttons/switch21.wav";
	}
	if (self->sounds == 2)
	{
		G_PrecacheSound ("buttons/switch02.wav");
		self->noise = "buttons/switch02.wav";
	}
	if (self->sounds == 3)
	{
		G_PrecacheSound ("buttons/switch04.wav");
		self->noise = "buttons/switch04.wav";
	}
	
	G_SetMovedir (self->angles, self->movedir);

	self->movetype = MOVETYPE_PUSH;
	self->solid = SOLID_BSP;
	gi.setmodel (self, self->model);
	G_LinkEntity (self);

	self->blocked = button_blocked;
	self->use = button_use;

	if (self->health)
	{
		self->max_health = self->health;
		self->die = button_killed;
		self->takedamage = DAMAGE_YES;
	}
	else
		self->touch = button_touch;

	if (!self->speed)
		self->speed = 40;
	if (!self->wait)
		self->wait = 1;
	if (!self->lip)
		self->lip = 4;

	self->state = STATE_BOTTOM;

	Vec3 size = self->maxs - self->mins;

	self->pos1 = self->origin;
	self->pos2 = self->pos1 + self->movedir*(fabs(self->movedir*size) - self->lip);
}
